/*
 * base64		smaller encoder/decoder for base64.
 *
 * Copyright 1995 (C) Olaf Kirch, <okir@monad.swb.de>
 *
 * This program is covered by the GNU GPL version 2.
 */
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <unistd.h>

static unsigned char *	itoa64 = 
"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

static unsigned char	a64toi[UCHAR_MAX+1];

#define TO64(i)		itoa64[(i) & 0x3F]
#define FROM64(c)	a64toi[c]

int
to64(FILE *ofp, unsigned char *from, int n)
{
	static unsigned char	temp[3];
	static unsigned char	to[61];
	static int		i = 0, j = 0;

	/* flush remainder */
	if (from == NULL) {
		while (i < 3)
			temp[i++] = 0;
		n = 0;
	}

	while (1) {
		for (i = 0; i < 3 && n > 0; i++, n--) {
			temp[i] = *from++;
		}
		if (i < 3)
			break; 	/* go back for more */

		to[j++] = TO64((temp[0] >> 2));
		to[j++] = TO64(((temp[0] << 4) & 0xF0) |
					((temp[1] >> 4) & 0x0F));
		to[j++] = TO64(((temp[1] << 2) & 0xFC) |
					((temp[2] >> 6) & 0x03));
		to[j++] = TO64(temp[2]);

		if (j == 60 || from == NULL) {
			to[j] = '\0';
			fprintf(ofp, "%s%s", to, from? "\n" : "");
			j = 0;
		}
	}
	return 1;
}

int
from64(FILE *ofp, unsigned char *from, int n)
{
	unsigned char	out[46];
	unsigned int	temp[4];
	int		i = 0, j;

	if (n > 60) {
		fprintf(stderr, "bad line: too long (%d chars)\n", n);
		return 0;
	}
	while (n && *from != '=') {
		for (j = 0; j < 4 && n > 0 && *from != '='; n--) {
			if ((temp[j++] = FROM64(*from++)) == 0xFF) {
				fprintf(stderr, "illegal character %c (%02x)\n",
							from[-1], from[-1]);
				return 0;
			}
		}
		while (j < 4)
			temp[j++] = 0;
		out[i++] = (temp[0] << 2) | ((temp[1] >> 4) & 0x0F);
		out[i++] = (temp[1] << 4) | ((temp[2] >> 2) & 0x3F);
		out[i++] = (temp[2] << 6) | temp[3];
	}
	out[i] = '\0';
	fwrite(out, 1, i, ofp);
	return 1;
}

int
base64(int decode, char *fname, FILE *ifp, FILE *ofp)
{
	unsigned char	buffer[81];
	int		n, done = 0, line = 1;

	buffer[81] = '\0';
	if (decode) {
		while (!done && fgets(buffer, 80, ifp) != NULL) {
			if (strchr(buffer, '\n') != NULL)
				line++;
			n = strlen(buffer);
			if (buffer[n-1] == '\n')
				buffer[--n] = '\0';
			if (n >= 2 && !strcmp(buffer + n - 2, "==")) {
				done = 1;
			}
			if (!from64(ofp, buffer, n)) {
				fprintf(stderr,
					"conversion error in %s, line %d\n",
							fname, line-1);
				return 0;
			}
		}
	} else {
		while ((n = fread(buffer, 1, 80, ifp)) > 0) {
			if (!to64(ofp, buffer, n)) {
				fprintf(stderr,
					"format error in %s, line %d\n",
							fname, line-1);
				return 0;
			}
		}
		if (!to64(ofp, NULL, 0)) {
			fprintf(stderr, "format error in %s, line %d\n",
						fname, line);
			return 0;
		}
		fprintf(ofp, "==\n");
	}
	return 1;
}

int
main(int argc, char **argv)
{
	FILE	*ifp, *ofp;
	int	decode = 0, i;

	argc--;
	argv++;
	while (argc > 0 && **argv == '-') {
		switch ((*argv)[1]) {
		case 'd':
			decode = 1;
			break;
		default:
			fprintf(stderr, "base64: unknown option %s\n", *argv);
			exit(2);
		}
		argc--;
		argv++;
	}

	for (i = 0; i < UCHAR_MAX; i++) {
		a64toi[i] = 0xFF;
	}
	for (i = 0; i < 64; i++) {
		a64toi[itoa64[i]] = i;
	}

	if (argc > 0) {
		while (argc > 0) {
			char	tempfile[PATH_MAX];

			if ((ifp = fopen(*argv, "r")) == NULL) {
				fprintf(stderr, "Can't open %s for reading",
							*argv);
				perror("");
				exit (2);
			}
			sprintf(tempfile, "%s.%d", *argv, getpid());
			if ((ofp = fopen(tempfile, "w")) == NULL) {
				fprintf(stderr, "Can't open %s for writing",
							tempfile);
				perror("");
				exit (2);
			}
			if (!base64(decode, *argv, ifp, ofp)) {
				unlink (tempfile);
				exit (2);
			}
			fclose(ifp);
			if (rename(tempfile, *argv) < 0) {
				fprintf(stderr, "Couldn't rename %s to %s",
							tempfile, *argv);
				perror("");
				exit (2);
			}
		}
	} else {
		if (!base64(decode, "<stdin>", stdin, stdout)) 
			exit (2);
	}
	return 0;
}
